# -*- coding: utf-8 -*-
#
# Copyright 2019 Pietro Barbiero and Giovanni Squillero
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json
import os
import pickle
from collections import OrderedDict
from typing import Callable, Iterable, List, Tuple

import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.pipeline import Pipeline
from sklearn.utils.validation import check_X_y

from .config import create_model_stmt, insert_model_stmt, query_model_stmt
from .database import _save_to_db, _load_from_db


class LazyPipeline(Pipeline):
    """
    A LazyPipeline estimator.

    A lazy pipeline is a sklearn-like pipeline that follows the memoization paradigm.
    Once the pipeline has been fitted, its steps are pickled and stored in a local
    database. Therefore, when the program starts again, the pipeline will fetch its fitted
    steps from the database and will skip the fit operation.

    Parameters
    ----------
    steps
        List of (name, transform) tuples (implementing fit/transform) that are
        chained, in the order in which they are chained, with the last object
        an estimator.

    database
        Used to cache the fitted transformers of the pipeline.
        It is the path to the database directory.
        Caching the transformers is advantageous when fitting is time consuming.

    verbose : bool, default=False
        If True, the time elapsed while fitting each step will be printed as it
        is completed.

    Attributes
    ----------
    named_steps : bunch object, a dictionary with attribute access
        Read-only attribute to access any step parameter by user given name.
        Keys are step names and values are steps parameters.

    See Also
    --------
    `sklearn.pipeline.Pipeline <https://scikit-learn.org/stable/modules/generated/sklearn.pipeline.Pipeline.html>`_

    Examples
    --------
    >>> from sklearn import svm
    >>> from sklearn.datasets import make_classification
    >>> from sklearn.feature_selection import SelectKBest
    >>> from sklearn.feature_selection import f_regression
    >>> from lazygrid.lazy_estimator import LazyPipeline
    >>> import pandas as pd
    >>> # generate some data to play with
    >>> X, y = make_classification(
    ...     n_informative=5, n_redundant=0, random_state=42)
    >>> X = pd.DataFrame(X)
    >>> # ANOVA SVM-C
    >>> anova_filter = SelectKBest(f_regression, k=5)
    >>> clf = svm.SVC(kernel='linear')
    >>> anova_svm = LazyPipeline([('anova', anova_filter), ('svc', clf)])
    >>> # You can set the parameters using the names issued
    >>> # For instance, fit using a k of 10 in the SelectKBest
    >>> # and a parameter 'C' of the svm
    >>> anova_svm.set_params(anova__k=10, svc__C=.1).fit(X, y)
    Pipeline(steps=[('anova', SelectKBest(...)), ('svc', SVC(...))])
    >>> prediction = anova_svm.predict(X)
    >>> anova_svm.score(X, y)
    0.83
    >>> # getting the selected features chosen by anova_filter
    >>> anova_svm['anova'].get_support()
    array([False, False,  True,  True, False, False,  True,  True, False,
           True, False,  True,  True, False,  True, False,  True,  True,
           False, False])
    >>> # Another way to get selected features chosen by anova_filter
    >>> anova_svm.named_steps.anova.get_support()
    array([False, False,  True,  True, False, False,  True,  True, False,
           True, False,  True,  True, False,  True, False,  True,  True,
           False, False])
    >>> # Indexing can also be used to extract a sub-pipeline.
    >>> sub_pipeline = anova_svm[:1]
    >>> sub_pipeline
    Pipeline(steps=[('anova', SelectKBest(...))])
    >>> coef = anova_svm[-1].coef_
    >>> anova_svm['svc'] is anova_svm[-1]
    True
    >>> coef.shape
    (1, 10)
    >>> sub_pipeline.inverse_transform(coef).shape
    (1, 20)
    """

    def __init__(self, steps, database: str = "./database/", verbose: bool = False):
        self.database = database
        super().__init__(steps, memory=None, verbose=verbose)

    # TODO: fit_params sould be checked
    def fit(self, X: pd.DataFrame, y: Iterable = None, **fit_params):
        """Fit the model

        Fit all the transforms one after the other and transform the
        data, then fit the transformed data using the final estimator.

        Parameters
        ----------
        X
            Training data. Must fulfill input requirements of first step of the
            pipeline.

        y
            Training targets. Must fulfill label requirements for all steps of
            the pipeline.

        **fit_params : dict of string -> object
            Parameters passed to the ``fit`` method of each step, where
            each parameter name is prefixed such that parameter ``p`` for step
            ``s`` has key ``s__p``.

        Returns
        -------
        self
            This estimator
        """

        assert isinstance(X, pd.DataFrame)

        # Check that X and y have correct shape
        Xnp, ynp = check_X_y(X, y)
        self._validate_steps()

        self.train_ = tuple(X.index)
        self.features_ = tuple(X.columns)
        self.database_ = os.path.join(self.database, "database.sqlite")
        self.fit_params_ = fit_params
        self._fit(X, y, **fit_params)
        self.is_fitted_ = True

        return self

    def _fit(self, X: pd.DataFrame, y: Iterable = None, **fit_params):
        Xt = X
        ids = ()
        # fit or load intermediate steps
        for (step_idx, name, transformer) in self._iter(with_final=False, filter_passthrough=False):
            transformer, ids, Xt = self._fit_step(transformer, ids, False, Xt, y, **fit_params)
            self.steps[step_idx] = (name, copy.deepcopy(transformer))

        # fit or load final step
        transformer, ids, Xt = self._fit_step(self.steps[-1][1], ids, True, Xt, y, **fit_params)
        self.steps[-1] = (self.steps[-1][0], copy.deepcopy(transformer))

        return self

    def _fit_step(self, transformer: BaseEstimator, ids: Tuple, is_final: bool,
                  X: pd.DataFrame, y: Iterable = None, **fit_params):
        # make transformer unique for each CV split
        transformer.train_ = tuple(X.index)
        transformer.features_ = tuple(X.columns)

        # load transformer from database
        transformer_loaded, ids_loaded = self._load(transformer, ids)
        is_loaded = False if transformer_loaded is None else True
        if is_loaded:
            transformer = transformer_loaded
            ids = ids_loaded

        # fit final step
        if is_final:
            if not is_loaded:
                transformer.fit(X, y, **fit_params)

        # fit intermediate steps
        else:
            if not is_loaded:
                transformer.fit(X, y, **fit_params)

            transformed_data = transformer.transform(X)

            if isinstance(transformed_data, Tuple):
                X, y = transformed_data

            else:
                Xnp = transformed_data

                # reshape input data
                if Xnp.shape != X.shape:
                    if isinstance(X, pd.DataFrame):
                        X = X.iloc[:, transformer.get_support()]

                else:
                    X = pd.DataFrame(Xnp)

        # save transformer
        if not is_loaded:
            ids = self._save(transformer, ids)

        return transformer, ids, X

    def _save(self, transformer: BaseEstimator, ids: Tuple):
        query, entry = _step_db(transformer, ids)
        result = _save_to_db(self.database_, entry, query, create_model_stmt, insert_model_stmt, query_model_stmt)
        if result:
            ids = ids + (result[0],)
        return ids

    def _load(self, transformer: BaseEstimator, ids: Tuple):
        query, entry = _step_db(transformer, ids)
        result = _load_from_db(self.database_, query, create_model_stmt, query_model_stmt)
        if result:
            ids = ids + (result[0],)
            transformer = pickle.loads(result[5])
            transformer.is_fitted_ = True
            return transformer, ids
        else:
            return None, None


def _step_db(estimator: BaseEstimator, ids: Tuple):
    estimator.is_fitted_ = True
    # make a dictionary of parameters
    pms = ()
    estimator_params = estimator.get_params()
    for key in sorted(estimator_params.keys()):
        value = estimator_params[key]
        if isinstance(value, Callable):
            pms = pms + (key, value.__name__)

        if value == "warn":
            pms = pms + (key, 10)

        # discard parameters which are not json serializable
        try:
            json.dumps(value)
            pms = pms + (key, value)
        except TypeError:
            continue

    query = (
        json.dumps(estimator.train_),
        json.dumps(estimator.features_),
        json.dumps(pms),
        json.dumps(ids),
    )
    entry = (
        *query,
        pickle.dumps(estimator),
    )

    return query, entry



# -*- coding: utf-8 -*-
#
# Copyright 2019 Pietro Barbiero and Giovanni Squillero
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import glob
import traceback
from difflib import SequenceMatcher
import numpy as np
import openml
import pandas as pd
import os
from sklearn.impute import SimpleImputer
from sklearn.datasets import fetch_openml
from sklearn.preprocessing import LabelEncoder
import logging


def load_npy_dataset(path_x: str, path_y: str) -> (np.ndarray, np.ndarray, int):
    """
    Load npy data set.

    Parameters
    ----------
    path_x
        Path to data matrix
    path_y
        Path to data labels

    Returns
    -------
    Tuple
        Data matrix, data labels, and number of classes

    Examples
    --------
    >>> import os
    >>> from sklearn.datasets import make_classification
    >>> import numpy as np
    >>> import lazygrid as lg
    >>>
    >>> x, y = make_classification(random_state=42)
    >>>
    >>> path_x, path_y = "x.npy", "y.npy"
    >>> np.save(path_x, x)
    >>> np.save(path_y, y)
    >>>
    >>> x, y, n_classes = lg.datasets.load_npy_dataset(path_x, path_y)
    """
    try:
        x = np.load(path_x)
        y = np.load(path_y)
        n_classes = len(np.unique(y))

    except FileNotFoundError:
        logging.exception("Exception occurred")
        return None, None, None

    return x, y, n_classes


def load_openml_dataset(data_id: int = None, dataset_name: str = None) -> (np.ndarray, np.ndarray, int):
    """
    Load OpenML data set.

    Parameters
    ----------
    data_id
        Data set identifier
    dataset_name
        Data set name

    Returns
    -------
    Tuple
        Data matrix, data labels, and number of classes

    Examples
    --------

    >>> import lazygrid as lg
    >>>
    >>> x, y, n_classes = lg.datasets.load_openml_dataset(dataset_name="iris")
    >>> n_classes
    3
    """
    assert (data_id is not None) or (dataset_name is not None)
    assert isinstance(dataset_name, str) or dataset_name is None

    try:
        if data_id is not None:
            x, y = fetch_openml(data_id=data_id, return_X_y=True)
        else:
            x, y = fetch_openml(name=dataset_name, return_X_y=True)

        if isinstance(x, pd.DataFrame):
            x = x.values

        if not isinstance(x, np.ndarray):
            x = x.toarray()

        si = SimpleImputer(missing_values=np.nan, strategy='mean')
        x = si.fit_transform(x)

        le = LabelEncoder()
        y = le.fit_transform(y)
        n_classes = np.unique(y).shape[0]

        return x, y, n_classes

    except Exception:
        logging.exception("Exception occurred")
        return [None, None, None]


# def _similar(a: str, b: str) -> float:
#     """
#     Compute how much two strings are similar.
#     """
#     return SequenceMatcher(None, a, b).ratio()


def _is_correct_task(task: str, db: dict) -> bool:
    """
    Check if the current data set is compatible with the specified task.

    Parameters
    ----------
    task
        Regression or classification
    db
        OpenML data set dictionary

    Returns
    -------
    bool
        True if the task and the data set are compatible
    """
    if task == "classification":
        return db['NumberOfSymbolicFeatures'] == 1 and db['NumberOfClasses'] > 0
    elif task == "regression":
        return True
    else:
        return False


def _load_datasets(output_dir: str = "./data", min_classes: int = 0, task: str = "classification",
                   max_samples: int = np.inf, max_features: int = np.inf) -> pd.DataFrame:
    """
    Load all OpenML data sets compatible with the requirements and save a .csv file as a reference.

    Parameters
    ----------
    output_dir
        Directory where the .csv file will be stored
    min_classes
        Minimum number of classes required for each data set
    task
        Classification or regression
    max_samples
        Maximum number of samples required for each data set
    max_features
        Maximum number of features required for each data set

    Returns
    -------
    Dataframe
        Information required to load the latest version of each data set

    Examples
    --------
    >>> datasets = _load_datasets(task="classification", min_classes=2, max_samples=1000, max_features=10)
    >>> datasets.loc["iris"]
    version          45
    did           42098
    n_samples       150
    n_features        4
    n_classes         3
    Name: iris, dtype: int64
    """

    data = {}

    for key, db in openml.datasets.list_datasets().items():

        try:
            logging.info("Loading data set: %s, ID: %d..." % (db['name'], db['did']))

            if db['NumberOfClasses'] > min_classes and _is_correct_task(task, db) and \
                    db['NumberOfInstances'] < max_samples and db['NumberOfFeatures'] < max_features and \
                    db['status'] == 'active':

                if db['name'] in data:
                    if db['version'] < data[db['name']]['version']:
                        continue

                # load_openml_dataset(db['did'])

                data[db['name']] = {}
                data[db['name']]['version'] = db['version']
                data[db['name']]['did'] = db['did']
                data[db['name']]['n_samples'] = db['NumberOfInstances']
                data[db['name']]['n_features'] = db['NumberOfNumericFeatures']
                data[db['name']]['n_classes'] = db['NumberOfClasses']

        except (IndexError, ValueError, KeyError):
            logging.info("Error loading data set: %s, ID: %d!" % (db['name'], db['did']))
            logging.info(traceback.format_exc())

    data = pd.DataFrame(data).transpose()
    try:
        data = data.sort_values(by=["n_samples", "n_features", "n_classes"]).astype('int64')
    except KeyError:
        logging.info(traceback.format_exc())
        return pd.DataFrame()

    if not os.path.isdir(output_dir):
        os.mkdir(output_dir)
    file_name = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S") + "-datalist.csv"
    data.to_csv(os.path.join(output_dir, file_name), index_label="db_name")

    return data


def fetch_datasets(output_dir: str = "./data", update_data: bool = False,
                   min_classes: int = 0, task: str = "classification",
                   max_samples: int = np.inf, max_features: int = np.inf) -> pd.DataFrame:
    """
    Load OpenML data sets compatible with the requirements.

    Parameters
    ----------
    output_dir
        Directory where the .csv file will be stored
    update_data
        If True it deletes cached data sets and downloads their latest version;
        otherwise it loads data sets as specified inside the cache
    min_classes
        Minimum number of classes required for each data set
    task
        Classification or regression
    max_samples
        Maximum number of samples required for each data set
    max_features
        Maximum number of features required for each data set

    Returns
    -------
    Dataframe
        Information required to load the latest version of each data set

    Examples
    --------
    >>> import lazygrid as lg
    >>>
    >>> datasets = lg.datasets.fetch_datasets(task="classification", min_classes=2, max_samples=1000, max_features=10)
    >>> datasets.loc["iris"]
    version          45
    did           42098
    n_samples       150
    n_features        4
    n_classes         3
    Name: iris, dtype: int64
    """
    files_location = os.path.join(output_dir, '*.csv')
    file_list = glob.glob(files_location)

    data = None

    # download (again) new data if necessary
    if not os.path.isdir(output_dir) or not file_list or update_data:
        data = _load_datasets(output_dir, min_classes, task, max_samples, max_features)

        # delete previous data files
        file_list.sort()
        for file in file_list[:-1]:
            os.remove(file)

    # fetch last version of downloaded data if there is one
    elif file_list:
        file_list = glob.glob(files_location)
        file_list.sort()
        data = pd.read_csv(file_list[-1], index_col=0).astype("int64")

    return data



# -*- coding: utf-8 -*-
#
# Copyright 2019 Pietro Barbiero and Giovanni Squillero
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sqlite3
from typing import Optional, Any, Iterable
import logging


def _save_to_db(db_name: str, entry: Iterable, query: Iterable,
                create_stmt: str, insert_stmt: str, query_stmt: str) -> Optional[Any]:
    """
    Save fitted model into a database.

    Parameters
    ----------
    db_name
        Database name
    entry
        Collection that will be inserted into the database
    query
        Collection required to fetch a database entry
    create_stmt
        Database statement for table creation
    insert_stmt
        Database statement for model insertion
    query_stmt
        Database statement for model query

    Returns
    -------
    Optional[Any]
        Query result
    """
    # # Sqlite does not accept INT larger than 8 bytes.
    # sqlite3.register_adapter(np.int64, lambda val: int(val))
    # sqlite3.register_adapter(np.int32, lambda val: int(val))

    # Connect to database if it exists
    root_dir = os.path.dirname(db_name)
    if not os.path.isdir(root_dir) and root_dir:
        os.makedirs(root_dir, exist_ok=True)
    db = sqlite3.connect(db_name)
    cursor = db.cursor()
    db.execute(create_stmt)

    # Insert item
    try:
        cursor.execute(insert_stmt, entry)
    except sqlite3.IntegrityError:
        # logging.exception("Exception occurred")
        pass
    result = cursor.execute(query_stmt, query).fetchone()

    db.commit()
    db.close()

    return result


def _load_from_db(db_name: str, query: Iterable, create_stmt: str, query_stmt: str) -> Optional[Any]:
    """
    Load fitted model from a database.

    Parameters
    ----------
    db_name
        Database name
    query
        Collection required to fetch a database entry
    create_stmt
        Database statement for table creation
    query_stmt
        Database statement for model query

    Returns
    -------
    Optional[Any]
        Query result
    """
    # # Sqlite does not accept INT larger than 8 bytes.
    # sqlite3.register_adapter(np.int64, lambda val: int(val))
    # sqlite3.register_adapter(np.int32, lambda val: int(val))

    # Connect to database if it exists
    root_dir = os.path.dirname(db_name)
    if not os.path.isdir(root_dir) and root_dir:
        os.makedirs(root_dir, exist_ok=True)
    db = sqlite3.connect(db_name)
    cursor = db.cursor()
    db.execute(create_stmt)

    result = cursor.execute(query_stmt, query).fetchone()
    cursor.close()

    return result


def load_all_from_db(db_name: str, table_name: str = "MODEL") -> Optional[Any]:
    """
    Load all database items.

    Parameters
    ----------
    db_name
        Database name
    table_name
        Database table to load

    Returns
    -------
    Optional[Any]
        Query result
    """
    # # Sqlite does not accept INT larger than 8 bytes.
    # sqlite3.register_adapter(np.int64, lambda val: int(val))
    # sqlite3.register_adapter(np.int32, lambda val: int(val))

    # Connect to database if it exists
    root_dir = os.path.dirname(db_name)
    if not os.path.isdir(root_dir) and root_dir:
        os.makedirs(root_dir, exist_ok=True)
    db = sqlite3.connect(db_name)
    cursor = db.cursor()

    stmt = '''SELECT name FROM sqlite_master WHERE type='table' AND name=? '''
    table = cursor.execute(stmt, (table_name,)).fetchone()
    if not table:
        db.close()
        return None

    query_stmt = '''SELECT * FROM %s''' % table_name
    result = cursor.execute(query_stmt).fetchall()
    cursor.close()

    return result


def drop_db(db_name: str) -> None:
    """
    Drop database table if it exists.

    Parameters
    ----------
    db_name
        Database name

    Returns
    -------
    None

    Examples
    --------
    >>> import lazygrid as lg
    >>>
    >>> lg.database.drop_db(db_name="my-database.sqlite")
    """
    root_dir = os.path.dirname(db_name)
    if not os.path.isdir(root_dir) and root_dir:
        os.makedirs(root_dir, exist_ok=True)
    db = sqlite3.connect(db_name)
    cursor = db.cursor()

    cursor.execute("DROP TABLE IF EXISTS MODEL")

    db.close()

    return




# -*- coding: utf-8 -*-
#
# Copyright 2019 Pietro Barbiero and Giovanni Squillero
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.

import matplotlib
from sklearn.model_selection import learning_curve
matplotlib.use('Agg')
import os
from typing import List
import numpy as np
import matplotlib.pyplot as plt


def plot_boxplots(scores: List, labels: List[str], file_name: str, title: str, output_dir: str = "./figures") -> dict:
    """
    Generate and save boxplots.

    Parameters
    ----------
    scores
        List of scores to compare
    labels
        Name / identifier of each score list
    file_name
        Output file name
    title
        Figure title
    output_dir
        Output directory

    Returns
    -------
    Boxplot
        boxplot object

    """
    file_name = os.path.join(output_dir, "box_plot_" + file_name + ".png")
    if not os.path.isdir(output_dir):
        os.mkdir(output_dir)

    cv = np.stack(scores, axis=1)

    plt.figure()
    results = plt.boxplot(cv, notch=True, labels=labels)
    plt.title(title)
    plt.tight_layout()
    plt.savefig(file_name, dpi=800)
    plt.show()

    return results


def plot_learning_curve(estimator, title, X, y, axes=None, ylim=None, cv=None,
                        n_jobs=None, train_sizes=np.linspace(.1, 1.0, 5)):
    """
    Generate 3 plots: the test and training learning curve, the training
    samples vs fit times curve, the fit times vs score curve.

    Parameters
    ----------
    estimator : object type that implements the "fit" and "predict" methods
        An object of that type which is cloned for each validation.

    title : string
        Title for the chart.

    X : array-like, shape (n_samples, n_features)
        Training vector, where n_samples is the number of samples and
        n_features is the number of features.

    y : array-like, shape (n_samples) or (n_samples, n_features), optional
        Target relative to X for classification or regression;
        None for unsupervised learning.

    axes : array of 3 axes, optional (default=None)
        Axes to use for plotting the curves.

    ylim : tuple, shape (ymin, ymax), optional
        Defines minimum and maximum yvalues plotted.

    cv : int, cross-validation generator or an iterable, optional
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:
          - None, to use the default 5-fold cross-validation,
          - integer, to specify the number of folds.
          - :term:`CV splitter`,
          - An iterable yielding (train, test) splits as arrays of indices.

        For integer/None inputs, if ``y`` is binary or multiclass,
        :class:`StratifiedKFold` used. If the estimator is not a classifier
        or if ``y`` is neither binary nor multiclass, :class:`KFold` is used.

        Refer :ref:`User Guide <cross_validation>` for the various
        cross-validators that can be used here.

    n_jobs : int or None, optional (default=None)
        Number of jobs to run in parallel.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.

    train_sizes : array-like, shape (n_ticks,), dtype float or int
        Relative or absolute numbers of training examples that will be used to
        generate the learning curve. If the dtype is float, it is regarded as a
        fraction of the maximum size of the training set (that is determined
        by the selected validation method), i.e. it has to be within (0, 1].
        Otherwise it is interpreted as absolute sizes of the training sets.
        Note that for classification the number of samples usually have to
        be big enough to contain at least one sample from each class.
        (default: np.linspace(0.1, 1.0, 5))
    """
    if axes is None:
        _, axes = plt.subplots(1, 2, figsize=(15, 5))

    axes[0].set_title(title)
    if ylim is not None:
        axes[0].set_ylim(*ylim)
    axes[0].set_xlabel("Training examples")
    axes[0].set_ylabel("Score")

    train_sizes, train_scores, test_scores, fit_times, _ = \
        learning_curve(estimator, X, y, cv=cv, n_jobs=n_jobs,
                       train_sizes=train_sizes, return_times=True)
    train_scores_mean = np.mean(train_scores, axis=1)
    train_scores_std = np.std(train_scores, axis=1)
    test_scores_mean = np.mean(test_scores, axis=1)
    test_scores_std = np.std(test_scores, axis=1)
    fit_times_mean = np.mean(fit_times, axis=1)
    fit_times_std = np.std(fit_times, axis=1)

    # Plot learning curve
    axes[0].grid()
    axes[0].fill_between(train_sizes, train_scores_mean - train_scores_std,
                         train_scores_mean + train_scores_std, alpha=0.1,
                         color="r")
    axes[0].fill_between(train_sizes, test_scores_mean - test_scores_std,
                         test_scores_mean + test_scores_std, alpha=0.1,
                         color="g")
    axes[0].plot(train_sizes, train_scores_mean, 'o-', color="r",
                 label="Training score")
    axes[0].plot(train_sizes, test_scores_mean, 'o-', color="g",
                 label="Cross-validation score")
    axes[0].legend(loc="best")

    # Plot n_samples vs fit_times
    axes[1].grid()
    axes[1].plot(train_sizes, fit_times_mean, 'o-')
    axes[1].fill_between(train_sizes, fit_times_mean - fit_times_std,
                         fit_times_mean + fit_times_std, alpha=0.1)
    axes[1].set_xlabel("Training examples")
    axes[1].set_ylabel("fit_times")
    axes[1].set_title("Scalability of the model")

    # # Plot fit_time vs score
    # axes[2].grid()
    # axes[2].plot(fit_times_mean, test_scores_mean, 'o-')
    # axes[2].fill_between(fit_times_mean, test_scores_mean - test_scores_std,
    #                      test_scores_mean + test_scores_std, alpha=0.1)
    # axes[2].set_xlabel("fit_times")
    # axes[2].set_ylabel("Score")
    # axes[2].set_title("Performance of the model")

    return plt



# -*- coding: utf-8 -*-
#
# Copyright 2019 Pietro Barbiero and Giovanni Squillero
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import traceback
from typing import Tuple, List, Iterator, Iterable
import copy
from keras import Model
from keras.wrappers.scikit_learn import KerasClassifier
from sklearn.pipeline import Pipeline
from sklearn.ensemble import RandomForestClassifier
from sklearn.svm import SVC
from sklearn.feature_selection import SelectKBest, f_classif
from sklearn.preprocessing import RobustScaler, StandardScaler
from itertools import chain, combinations, product
from .lazy_estimator import LazyPipeline


def _powerset(iterable: Iterable) -> Iterator:
    """
    Compute the powerset of a collection.

    Parameters
    ----------
    iterable
        Collection of items

    Returns
    -------
    Iterator
        Powerset of the collection
    """
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s)+1))


def generate_grid(elements: list, lazy: bool = True, **kwargs) -> list:
    """
    Generate all possible combinations of sklearn Pipelines given the input steps.

    Parameters
    ----------
    elements
        List of elements used to generate the pipelines
    lazy
        If True it generates LazyPipelines objects;
        if False it generates standard sklearn Pipeline objects
    kwargs
        Keyword arguments to generate Pipeline objects

    Returns
    -------
    list
        List of pipelines

    Example
    --------
    >>> from sklearn.ensemble import RandomForestClassifier
    >>> from sklearn.svm import SVC
    >>> from sklearn.feature_selection import SelectKBest, f_classif
    >>> from sklearn.preprocessing import RobustScaler, StandardScaler
    >>> import lazygrid as lg
    >>>
    >>> preprocessors = [StandardScaler(), RobustScaler()]
    >>> feature_selectors = [SelectKBest(score_func=f_classif, k=1), SelectKBest(score_func=f_classif, k=2)]
    >>> classifiers = [RandomForestClassifier(random_state=42), SVC(random_state=42)]
    >>>
    >>> elements = [preprocessors, feature_selectors, classifiers]
    >>>
    >>> pipelines = lg.grid.generate_grid(elements)
    """

    assert isinstance(elements, list)
    assert all([isinstance(step, list) for step in elements])

    # generate all possible combinations of steps
    subsets = []
    for step_list in product(*elements):
        for steps in _powerset(step_list):
            if len(steps) > 0:
                if steps[-1] == step_list[-1]:
                    subsets.append(steps)

    pipelines = []
    for subset in subsets:
        # create list of tuples (step_name, step_object) to feed sklearn Pipeline
        i = 0
        steps = []
        for step in subset:
            steps.append(("step_" + str(i), copy.deepcopy(step)))
            i += 1

        if lazy:
            pipeline = LazyPipeline(steps, **kwargs)
        else:
            pipeline = Pipeline(steps, **kwargs)

        pipelines.append(pipeline)

    return pipelines


def generate_grid_search(model: KerasClassifier, model_params: dict,
                         fit_params: dict) -> Tuple[List[Model], List[dict]]:
    """
    Generate all possible combinations of models.

    Parameters
    ----------
    model
        Model architecture
    model_params
        Model parameters. For each key the dictionary should contain a list of possible values
    fit_params
        Fit parameters. For each key the dictionary should contain a list of possible values

    Returns
    -------
    Tuple
        Models and their corresponding fit parameters

    Example
    --------
    >>> import keras
    >>> from keras import Sequential
    >>> from keras.layers import Conv2D, MaxPooling2D, Flatten, Dense
    >>> import lazygrid as lg
    >>> from keras.wrappers.scikit_learn import KerasClassifier
    >>>
    >>> # define keras model generator
    >>> def create_keras_model(input_shape, optimizer, n_classes):
    ...     kmodel = Sequential()
    ...     kmodel.add(Conv2D(32, kernel_size=(5, 5), strides=(1, 1), activation='relu', input_shape=input_shape))
    ...     kmodel.add(MaxPooling2D(pool_size=(2, 2)))
    ...     kmodel.add(Flatten())
    ...     kmodel.add(Dense(1000, activation='relu'))
    ...     kmodel.add(Dense(n_classes, activation='softmax'))
    ...
    ...     kmodel.compile(loss=keras.losses.categorical_crossentropy,
    ...                    optimizer=optimizer, metrics=['accuracy'])
    ...     return kmodel
    >>>
    >>> # cast keras model into sklearn model
    >>> kmodel = KerasClassifier(create_keras_model)
    >>>
    >>> # define all possible model parameters of the grid
    >>> model_params = {"optimizer": ['SGD', 'RMSprop'], "input_shape": [(28, 28, 3)], "n_classes": [10]}
    >>> fit_params = {"epochs": [5, 10, 20], "batch_size": [10, 20]}
    >>>
    >>> # generate all possible models given the parameters' grid
    >>> models, fit_parameters = lg.grid.generate_grid_search(kmodel, model_params, fit_params)
    """

    models = []
    fit_parameters = []
    keys = []
    values = []
    is_for_fit = []

    # fill keys' and values' lists with model parameters
    for key_model, value_model in model_params.items():
        keys.append(key_model)
        values.append(value_model)
        is_for_fit.append(False)

    # fill keys' and values' lists with fit parameters
    for key_fit, value_fit in fit_params.items():
        keys.append(key_fit)
        values.append(value_fit)
        is_for_fit.append(True)

    # generate all possible combinations of parameters
    for values_list in product(*values):

        learner = copy.deepcopy(model)

        # uniquely define model structure
        model_params_values = [values_list[i] for i in range(0, len(values_list)) if is_for_fit[i] is False]
        model_params_keys = [keys[i] for i in range(0, len(keys)) if is_for_fit[i] is False]
        model_params_dict = dict(zip(model_params_keys, model_params_values))
        learner = learner.build_fn(**model_params_dict)

        # uniquely define fit function
        fit_params_values = [values_list[i] for i in range(0, len(values_list)) if is_for_fit[i] is True]
        fit_params_keys = [keys[i] for i in range(0, len(keys)) if is_for_fit[i] is True]
        fit_params_dict = dict(zip(fit_params_keys, fit_params_values))
        functools.partial(learner.fit, **fit_params_dict)

        models.append(learner)
        fit_parameters.append(fit_params_dict)

    return models, fit_parameters



# -*- coding: utf-8 -*-
#
# Copyright 2019 Pietro Barbiero and Giovanni Squillero
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.

create_model_stmt = '''CREATE TABLE IF NOT EXISTS MODEL(
        id INTEGER PRIMARY KEY,
        train TEXT NOT NULL,
        features TEXT NOT NULL,
        parameters TEXT NOT NULL,
        ids TEXT NOT NULL,
        estimator BLOB NOT NULL,

        UNIQUE (train, features, parameters, ids)
        )'''

insert_model_stmt = '''INSERT INTO MODEL(
        train, features, parameters, ids, estimator)
        VALUES(?, ?, ?, ?, ?)'''

query_model_stmt = '''SELECT * FROM MODEL
                      WHERE train=? AND features=? AND parameters=? AND ids=?'''



# -*- coding: utf-8 -*-
#
# Copyright 2019 Pietro Barbiero and Giovanni Squillero
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from typing import Callable, List
from scipy import stats
from scipy.stats import mannwhitneyu
from sklearn.datasets import make_classification
from sklearn.linear_model import RidgeClassifier, LogisticRegression
from sklearn.metrics import make_scorer, recall_score, f1_score, roc_auc_score, accuracy_score, confusion_matrix


def confidence_interval_mean_t(x: np.ndarray, cl: float = 0.05) -> List:
    """
    Compute the confidence interval of the mean from sample data.

    Parameters
    ----------
    x
        Sample
    cl
        Confidence level

    Returns
    -------
    List
        confidence interval

    Examples
    --------
    >>> import numpy as np
    >>> import lazygrid as lg
    >>>
    >>> np.random.seed(42)
    >>> x = np.random.normal(loc=0, scale=2, size=10)
    >>> confidence_level = 0.05
    >>>
    >>> lg.statistics.confidence_interval_mean_t(x, confidence_level)
    [-0.13829578539063092, 1]


    Notes
    -----
    You should use the t distribution rather than the normal distribution
    when the variance is not known and has to be estimated from sample data.

    When the sample size is large, say 100 or above, the t distribution
    is very similar to the standard normal distribution.
    However, with smaller sample sizes, the t distribution is leptokurtic,
    which means it has relatively more scores in its tails than does the normal distribution.
    As a result, you have to extend farther from the mean to contain a given proportion of the area.
    """
    if np.all(x == np.mean(x)):
        return [np.mean(x), np.mean(x)]
    bounds = stats.t.interval(1-cl, len(x)-1, loc=np.mean(x), scale=stats.sem(x))
    adjusted_bounds = [bound if bound <= 1 else 1 for bound in bounds]
    return adjusted_bounds


def find_best_solution(solutions: list,
                       test: Callable = mannwhitneyu,
                       alpha: float = 0.05,
                       **kwargs) -> (int, list, list):
    """
    Find the best solution in a list of candidates, according to
    a statistical test and a significance level (alpha).

    The best solution is defined as the one having the highest mean value.

    Parameters
    ----------
    solutions
        List of candidate solutions
    test
        Statistical test
    alpha
        Significance level
    kwargs
        Keyword arguments required by the statistical test

    Returns
    -------
    Tuple
        - the position of the best solution inside the candidate input list;
        - the positions of the solutions which are not separable from the best one;
        - the list of p-values returned by the statistical test while comparing the best solution to the other candidates

    Examples
    --------
    >>> from sklearn.linear_model import LogisticRegression, RidgeClassifier
    >>> from sklearn.ensemble import RandomForestClassifier
    >>> from sklearn.datasets import make_classification
    >>> from sklearn.model_selection import cross_val_score
    >>> import lazygrid as lg
    >>>
    >>> x, y = make_classification(random_state=42)
    >>>
    >>> model1 = LogisticRegression(random_state=42)
    >>> model2 = RandomForestClassifier(random_state=42)
    >>> model3 = RidgeClassifier(random_state=42)
    >>> model_names = ["LogisticRegression", "RandomForestClassifier", "RidgeClassifier"]
    >>>
    >>> score1 = cross_val_score(estimator=model1, X=x, y=y, cv=10)
    >>> score2 = cross_val_score(estimator=model2, X=x, y=y, cv=10)
    >>> score3 = cross_val_score(estimator=model3, X=x, y=y, cv=10)
    >>>
    >>> scores = [score1, score2, score3]
    >>> best_idx, best_solutions_idx, pvalues = lg.statistics.find_best_solution(scores)
    >>> model_names[best_idx]
    'LogisticRegression'
    >>> best_solutions_idx
    [0, 2]
    >>> pvalues #doctest: +ELLIPSIS
    [0.4782..., 0.0360..., 0.1610...]
    """

    best_idx = 0
    best_solution = solutions[best_idx]
    best_mean = np.mean(best_solution)

    # find the best solution (the one having the highest mean value)
    index = 1
    for solution in solutions[index:]:
        solution_mean = np.mean(solution)

        if solution_mean > best_mean:
            best_solution = solution
            best_mean = np.mean(best_solution)
            best_idx = index

        index += 1

    best_solutions_idx = []
    pvalues = []

    # check if there are other candidates which may be equivalent to the best one
    index = 0
    for solution in solutions:
        try:
            statistic, pvalue = test(best_solution, solution, **kwargs)
        except ValueError:
            statistic, pvalue = np.inf, np.inf

        if pvalue > alpha:
            best_solutions_idx.append(index)
        pvalues.append(pvalue)

        index += 1

    return best_idx, best_solutions_idx, pvalues


def tn(y_true, y_pred):
    return confusion_matrix(y_true, y_pred)[0, 0]


def fp(y_true, y_pred):
    return confusion_matrix(y_true, y_pred)[0, 1]


def fn(y_true, y_pred):
    return confusion_matrix(y_true, y_pred)[1, 0]


def tp(y_true, y_pred):
    return confusion_matrix(y_true, y_pred)[1, 1]


def specificity(y_true, y_pred):
    return tn(y_true, y_pred) / (tn(y_true, y_pred) + fp(y_true, y_pred))


def sensitivity(y_true, y_pred):
    return tp(y_true, y_pred) / (tp(y_true, y_pred) + fn(y_true, y_pred))


scoring_summary = {'tp': make_scorer(tp), 'tn': make_scorer(tn),
                   'fp': make_scorer(fp), 'fn': make_scorer(fn),
                   'specificity': make_scorer(specificity), 'sensitivity': make_scorer(sensitivity),
                   'recall': make_scorer(recall_score), 'accuracy': make_scorer(accuracy_score),
                   'f1': make_scorer(f1_score), 'roc_auc': make_scorer(roc_auc_score)}
